/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package net.java.html.canvas;

import net.java.html.canvas.Style.Color;
import net.java.html.canvas.Style.LinearGradient;
import net.java.html.canvas.Style.Pattern;
import net.java.html.canvas.Style.RadialGradient;
import net.java.html.canvas.spi.GraphicsEnvironment;
import org.apidesign.html.canvas.impl.CnvsAccssr;

/**
 * A 2D Graphics Context similar to HTML5 or JavaFX GraphicsContext. Use this to
 * paint on your Canvas.s
 *
 * @author antonepple
 */
public final class GraphicsContext {

    GraphicsEnvironment graphicsEnvironmentImpl;

    static {
        CnvsAccssr cnvsAccssr = new CnvsAccssr() {
            @Override
            public GraphicsContext create(GraphicsEnvironment environment) {
                return new GraphicsContext(environment);
            }
        };
    }

    GraphicsContext(GraphicsEnvironment graphicsEnvironment) {
        this.graphicsEnvironmentImpl = graphicsEnvironment;
    }

    /**
     * Adds path elements to the current path to make an arc.
     *
     * @param centerX the center x position of the arc.
     * @param centerY the center y position of the arc.
     * @param radius the radius of the arc.
     * @param endAngle teh endAngle of the arc
     * @param ccw the direction of the arc (counterclockwise)
     */
    public void arc(double centerX,
            double centerY,
            double startAngle,
            double radius,
            double endAngle,
            boolean ccw) {
        graphicsEnvironmentImpl.arc(centerX, centerY, startAngle, radius, endAngle, ccw);
    }

    /**
     * Adds segments to the current path to make an arc.
     *
     * @param x1 the X coordinate of the first point of the arc.
     * @param y1 the Y coordinate of the first point of the arc.
     * @param x2 the X coordinate of the second point of the arc.
     * @param y2 the Y coordinate of the second point of the arc.
     * @param radius the radius of the arc in the range {0.0-positive infinity}.
     */
    public void arcTo(double x1,
            double y1,
            double x2,
            double y2,
            double radius) {
        graphicsEnvironmentImpl.arcTo(x1, y1, x2, y2, radius);
    }

    /**
     * Returns true if the the given x,y point is inside the path.
     *
     * @param x the X coordinate to use for the check.
     * @param y the Y coordinate to use for the check.
     * @return true if the point given is inside the path, false otherwise.
     */
    public boolean isPointInPath(double x, double y) {
        return graphicsEnvironmentImpl.isPointInPath(x, y);
    }

    /**
     * Fills the path with the current fill paint.
     */
    public void fill() {
        graphicsEnvironmentImpl.fill();
    }

    /**
     * Strokes the path with the current stroke paint.
     */
    public void stroke() {
        graphicsEnvironmentImpl.stroke();
    }

    /**
     * Starts a Path
     */
    public void beginPath() {
        graphicsEnvironmentImpl.beginPath();
    }

    /**
     * Closes the path.
     */
    public void closePath() {
        graphicsEnvironmentImpl.closePath();
    }

    /**
     * Clips using the current path
     */
    public void clip() {
        graphicsEnvironmentImpl.clip();
    }

    /**
     * Issues a move command for the current path to the given x,y coordinate.
     *
     * @param x0 the X position for the move to command.
     * @param y0 the Y position for the move to command.
     */
    public void moveTo(double x, double y) {
        graphicsEnvironmentImpl.moveTo(x, y);
    }

    /**
     * Adds segments to the current path to make a line at the given x,y
     * coordinate.
     *
     * @param x1 the X coordinate of the ending point of the line.
     * @param y1 the Y coordinate of the ending point of the line.
     */
    public void lineTo(double x, double y) {
        graphicsEnvironmentImpl.lineTo(x, y);
    }

    /**
     * Adds segments to the current path to make a quadratic curve.
     *
     * @param cpx the X coordinate of the control point
     * @param cpy the Y coordinate of the control point
     * @param x the X coordinate of the end point
     * @param y the Y coordinate of the end point
     */
    public void quadraticCurveTo(double cpx, double cpy, double x, double y) {
        graphicsEnvironmentImpl.quadraticCurveTo(cpx, cpy, x, y);
    }

    /**
     * Adds segments to the current path to make a cubic bezier curve.
     *
     * @param cp1x the X coordinate of first bezier control point.
     * @param cp1y the Y coordinate of the first bezier control point.
     * @param cp2x the X coordinate of the second bezier control point.
     * @param cp2y the Y coordinate of the second bezier control point.
     * @param x the X coordinate of the end point.
     * @param y the Y coordinate of the end point.
     */
    public void bezierCurveTo(double cp1x, double cp1y, double cp2x, double cp2y, double x, double y) {
        graphicsEnvironmentImpl.bezierCurveTo(cp1x, cp1y, cp2x, cp2y, x, y);
    }

    /**
     * Fills a rectangle using the current fill paint.
     *
     * @param x the X position of the upper left corner of the rectangle.
     * @param y the Y position of the upper left corner of the rectangle.
     * @param w the width of the rectangle.
     * @param h the height of the rectangle.
     */
    public void fillRect(double x, double y, double width, double height) {
        graphicsEnvironmentImpl.fillRect(x, y, width, height);
    }

    /**
     * Strokes a rectangle using the current stroke paint.
     *
     * @param x the X position of the upper left corner of the rectangle.
     * @param y the Y position of the upper left corner of the rectangle.
     * @param width the width of the rectangle.
     * @param height the height of the rectangle.
     */
    public void strokeRect(double x, double y, double width, double height) {
        graphicsEnvironmentImpl.strokeRect(x, y, width, height);
    }

    /**
     * Clears a portion of the canvas with a transparent color value.
     *
     * @param x X position of the upper left corner of the rectangle.
     * @param y Y position of the upper left corner of the rectangle.
     * @param width width of the rectangle.
     * @param height height of the rectangle.
     */
    public void clearRect(double x, double y, double width, double height) {
        graphicsEnvironmentImpl.clearRect(x, y, width, height);
    }

    /**
     * Clears a portion of the canvas with a transparent color value.
     *
     * @param x X position of the upper left corner of the rectangle.
     * @param y Y position of the upper left corner of the rectangle.
     * @param width width of the rectangle.
     * @param height height of the rectangle.
     */
    public void rect(double x, double y, double width, double height) {
        graphicsEnvironmentImpl.rect(x, y, width, height);
    }

    /**
     * Saves the following attributes onto a stack.
     * <ul>
     * <li>Global Alpha</li>
     * <li>Global Blend Operation</li>
     * <li>Transform</li>
     * <li>Fill Paint</li>
     * <li>Stroke Paint</li>
     * <li>Line Width</li>
     * <li>Line Cap</li>
     * <li>Line Join</li>
     * <li>Miter Limit</li>
     * <li>Number of Clip Paths</li>
     * <li>Font</li>
     * <li>Text Align</li>
     * <li>Text Baseline</li>
     * <li>Effect</li>
     * <li>Fill Rule</li>
     * </ul>
     * This method does NOT alter the current state in any way. Also, not that
     * the current path is not saved.
     */
    public void save() {
        graphicsEnvironmentImpl.save();
    }

    /**
     * Pops the state off of the stack, setting the following attributes to
     * their value at the time when that state was pushed onto the stack. If the
     * stack is empty then nothing is changed.
     *
     * <ul>
     * <li>Global Alpha</li>
     * <li>Global Blend Operation</li>
     * <li>Transform</li>
     * <li>Fill Paint</li>
     * <li>Stroke Paint</li>
     * <li>Line Width</li>
     * <li>Line Cap</li>
     * <li>Line Join</li>
     * <li>Miter Limit</li>
     * <li>Number of Clip Paths</li>
     * <li>Font</li>
     * <li>Text Align</li>
     * <li>Text Baseline</li>
     * <li>Effect</li>
     * <li>Fill Rule</li>
     * </ul>
     */
    public void restore() {
        graphicsEnvironmentImpl.restore();
    }

    /**
     * Rotates the current transform in degrees.
     *
     * @param angle value in degrees to rotate the current transform.
     */
    public void rotate(double angle) {
        graphicsEnvironmentImpl.rotate(angle);
    }

    /**
     * Concatenates the input with the current transform.
     *
     * @param mxx - the X coordinate scaling element of the 3x4 matrix
     * @param myx - the Y coordinate shearing element of the 3x4 matrix
     * @param mxy - the X coordinate shearing element of the 3x4 matrix
     * @param myy - the Y coordinate scaling element of the 3x4 matrix
     * @param mxt - the X coordinate translation element of the 3x4 matrix
     * @param myt - the Y coordinate translation element of the 3x4 matrix
     */
    public void transform(double mxx, double myx, double mxy, double myy, double mxt, double myt) {
        graphicsEnvironmentImpl.transform(mxx, myx, mxy, myy, mxt, myt);
    }

    /**
     * Concatenates the input with the current transform.
     *
     * @param mxx - the X coordinate scaling element of the 3x4 matrix
     * @param myx - the Y coordinate shearing element of the 3x4 matrix
     * @param mxy - the X coordinate shearing element of the 3x4 matrix
     * @param myy - the Y coordinate scaling element of the 3x4 matrix
     * @param mxt - the X coordinate translation element of the 3x4 matrix
     * @param myt - the Y coordinate translation element of the 3x4 matrix
     */
    public void setTransform(double mxx, double myx, double mxy, double myy, double mxt, double myt) {
        graphicsEnvironmentImpl.setTransform(mxx, myx, mxy, myy, mxt, myt);
    }

    /**
     * Translates the current transform by x, y.
     *
     * @param x value to translate along the x axis.
     * @param y value to translate along the y axis.
     */
    public void translate(double x, double y) {
        graphicsEnvironmentImpl.translate(x, y);
    }

    /**
     * Scales the current transform by x, y.
     *
     * @param x value to scale in the x axis.
     * @param y value to scale in the y axis.
     */
    public void scale(double x, double y) {
        graphicsEnvironmentImpl.scale(x, y);
    }

    /**
     * Draws an image at the given x, y position using the width and height of
     * the given image.
     *
     * @param img the image to be drawn.
     * @param x the X coordinate on the destination for the upper left of the
     * image.
     * @param y the Y coordinate on the destination for the upper left of the
     * image.
     */
    public void drawImage(Image image, double x, double y) {
        Object nativeImage = graphicsEnvironmentImpl.drawImage(image, x, y, image.getCached());
        image.cache(nativeImage);
    }

    /**
     * Draws an image into the given destination rectangle of the canvas. The
     * Image is scaled to fit into the destination rectagnle.
     *
     * @param img the image to be drawn.
     * @param x the X coordinate on the destination for the upper left of the
     * image.
     * @param y the Y coordinate on the destination for the upper left of the
     * image.
     * @param width the width of the destination rectangle.
     * @param height the height of the destination rectangle.
     */
    public void drawImage(Image image, double x, double y, double width, double height) {
        Object nativeImage = graphicsEnvironmentImpl.drawImage(image, x, y, width, height, image.getCached());
        image.cache(nativeImage);
    }

    /**
     * Draws the current source rectangle of the given image to the given
     * destination rectangle of the Canvas.
     *
     * @param img the image to be drawn.
     * @param sx the source rectangle's X coordinate position.
     * @param sy the source rectangle's Y coordinate position.
     * @param sw the source rectangle's width.
     * @param sh the source rectangle's height.
     * @param dx the destination rectangle's X coordinate position.
     * @param dy the destination rectangle's Y coordinate position.
     * @param dw the destination rectangle's width.
     * @param dh the destination rectangle's height.
     */
    public void drawImage(Image image, double sx, double sy, double sw, double sh, double dx, double dy, double dw, double dh) {
        Object nativeImage = graphicsEnvironmentImpl.drawImage(image, sx, sy, sw, sh, dx, dy, dw, dh, image.getCached());
        image.cache(nativeImage);
    }

    /**
     * Merges two images drawing one on top of the other and returning the
     * result.
     *
     * @param a the lower Image
     * @param b the upper Image
     * @return
     */
    public Image merge(Image a, Image b) {
        if (a.getCached() == null) {
            drawImage(a, 0, 0);
        }
        if (b.getCached() == null) {
            drawImage(b, 0, 0);
        }
        Object nativeImage = graphicsEnvironmentImpl.mergeImages(a, b, a.getCached(), b.getCached());
        Image merged = Image.create("should add real path here");
        merged.cache(nativeImage);
        return merged;
    }

//    public void setShadowColor(String color) {
//        graphicsEnvironmentImpl.setShadowColor(color);
//    }
//
//    public void setShadowBlur(double blur) {
//        graphicsEnvironmentImpl.setShadowBlur(blur);
//    }
//
//    public void setShadowOffsetX(double x) {
//        graphicsEnvironmentImpl.setShadowOffsetX(x);
//    }
//
//    public void setShadowOffsetY(double y) {
//        graphicsEnvironmentImpl.setShadowOffsetY(y);
//    }
//
//    public String getShadowColor() {
//        return graphicsEnvironmentImpl.getShadowColor();
//    }
//
//    public double getShadowBlur() {
//        return graphicsEnvironmentImpl.getShadowBlur();
//    }
//
//    public double getShadowOffsetX() {
//        return graphicsEnvironmentImpl.getShadowOffsetX();
//    }
//
//    public double getShadowOffsetY() {
//        return graphicsEnvironmentImpl.getShadowOffsetY();
//    }
    public String getLineCap() {
        return graphicsEnvironmentImpl.getLineCap();
    }

    public void setLineCap(String style) {
        graphicsEnvironmentImpl.setLineCap(style);
    }

    public String getLineJoin() {
        return graphicsEnvironmentImpl.getLineJoin();
    }

    public void setLineJoin(String style) {
        graphicsEnvironmentImpl.setLineJoin(style);
    }

    public double getLineWidth() {
        return graphicsEnvironmentImpl.getLineWidth();
    }

    public void setLineWidth(double width) {
        graphicsEnvironmentImpl.setLineWidth(width);
    }

    public double getMiterLimit() {
        return graphicsEnvironmentImpl.getMiterLimit();
    }

    public void setMiterLimit(double limit) {
        graphicsEnvironmentImpl.setMiterLimit(limit);
    }

    public void setFillStyle(Style style) {
        Object nativeFillStyle = graphicsEnvironmentImpl.setFillStyle(style, style.getCached());
        style.cache(nativeFillStyle);
    }

    public String getFont() {
        return graphicsEnvironmentImpl.getFont();
    }

    public void setFont(String font) {
        graphicsEnvironmentImpl.setFont(font);
    }

    public void setStrokeStyle(Style style) {
        Object nativeStrokeStyle = graphicsEnvironmentImpl.setStrokeStyle(style, style.getCached());
        style.cache(nativeStrokeStyle);
    }

    public String getTextAlign() {
        return graphicsEnvironmentImpl.getTextAlign();
    }

    public void setTextAlign(String textAlign) {
        graphicsEnvironmentImpl.setTextAlign(textAlign);
    }

    public String getTextBaseline() {
        return graphicsEnvironmentImpl.getTextBaseline();
    }

    public void setTextBaseline(String textbaseline) {
        graphicsEnvironmentImpl.setTextBaseline(textbaseline);
    }

    public void fillText(String text, double x, double y) {
        graphicsEnvironmentImpl.fillText(text, x, y);
    }

    public void fillText(String text, double x, double y, double maxWidth) {
        graphicsEnvironmentImpl.fillText(text, x, y, maxWidth);
    }

    public Dimension measureText(String text) {
        return graphicsEnvironmentImpl.measureText(text);
    }

    public void strokeText(String text, double x, double y) {
        graphicsEnvironmentImpl.strokeText(text, x, y);
    }

    public void strokeText(String text, double x, double y, double maxWidth) {
        graphicsEnvironmentImpl.strokeText(text, x, y, maxWidth);
    }

//    public ImageData createPixelMap(double x, double y) {
//        return graphicsEnvironmentImpl.createPixelMap(x, y);
//    }
//
//    public ImageData createPixelMap(ImageData pixelMap) {
//        return graphicsEnvironmentImpl.createPixelMap(pixelMap);
//    }
//
//    public ImageData getSnapshot(double x, double y, double width, double height) {
//        return graphicsEnvironmentImpl.getPixelMap(x, y, width, height);
//    }
//
//    public void drawPixelMap(ImageData pixelMap, double x, double y) {
//        graphicsEnvironmentImpl.putPixelMap(pixelMap, x, y);
//    }
//
//    public void drawPixelMap(ImageData pixelMap, double x, double y, double dirtyx, double dirtyy, double dirtywidth, double dirtyheight) {
//        graphicsEnvironmentImpl.putPixelMap(pixelMap, x, y, dirtyx, dirtyy, dirtywidth, dirtyheight);
//    }
    /**
     * Sets the global alpha of the current state.
     *
     * @param alpha value in the range {@code 0.0-1.0}. The value is clamped if
     * it is out of range.
     */
    public void setGlobalAlpha(double alpha) {
        graphicsEnvironmentImpl.setGlobalAlpha(alpha);
    }

    /**
     * Gets the current global alpha.
     *
     * @return the current global alpha.
     */
    public double getGlobalAlpha() {
        return graphicsEnvironmentImpl.getGlobalAlpha();
    }

    /**
     * Sets the global blend mode.
     *
     * @param op the BlendMode that will be set.
     */
    public void setGlobalCompositeOperation(String operation) {
        graphicsEnvironmentImpl.setGlobalCompositeOperation(operation);
    }

    /**
     * Gets the global blend mode.
     *
     * @return the global BlendMode of the current state.
     */
    public String getGlobalCompositeOperation() {
        return graphicsEnvironmentImpl.getGlobalCompositeOperation();
    }

    public LinearGradient createLinearGradient(double x0, double y0, double x1, double y1) {
        return new Style.LinearGradient(x0, y0, x1, y1);
    }

    public Pattern createPattern(Image image, String repeat) {
        return new Pattern(image, repeat);
    }

    public RadialGradient createRadialGradient(double x0, double y0, double r0, double x1, double y1, double r1) {
        return new RadialGradient(x0, y0, r0, x1, y1, r1);
    }

    public Color getWebColor(String webColor) {
        return new Style.Color(webColor);
    }

    /**
     * Get the height of this GraphicsContext (which should be the same as the
     * enclosing canvas height)
     *
     * @return the height of this GraphicsContext
     */
    public int getHeight() {
        return graphicsEnvironmentImpl.getHeight();
    }

    /**
     * Get the width of this GraphicsContext (which should be the same as the
     * enclosing canvas height)
     *
     * @return the width of this GraphicsContext
     */
    public int getWidth() {
        return graphicsEnvironmentImpl.getWidth();
    }

//    public void setHeight(int height) {
//        graphicsEnvironmentImpl.setHeight(height);
//    }
//
//    public void setWidth(int width) {
//        graphicsEnvironmentImpl.setWidth(width);
//    }
    /**
     * Fill a circle with a center position of centerX, centerY and the
     * specified radius.
     *
     * @param centerX
     * @param centerY
     * @param radius
     */
    public void fillCircle(float centerX, float centerY, float radius) {
        graphicsEnvironmentImpl.arc(centerX, centerY, radius, 0, Math.PI * 2, false);
    }

    /**
     * Fills a polygon with the given points using the currently set fill paint.
     *
     * @param x_coord array containing the x coordinates of the polygon's
     * points.
     * @param y_coord array containing the y coordinates of the polygon's
     * points.
     * @param vertexCount the number of points that make the polygon.
     */
    public void fillPolygon(double[] x_coord, double[] y_coord, int vertexCount) {
        if (vertexCount >= 1 && x_coord != null && x_coord.length >= vertexCount && y_coord != null && y_coord.length >= vertexCount) {
            graphicsEnvironmentImpl.beginPath();
        }
        graphicsEnvironmentImpl.moveTo(x_coord[0], y_coord[0]);
        for (int i = 1; i < vertexCount; i++) {
            graphicsEnvironmentImpl.lineTo(x_coord[i], y_coord[i]);

        }
        graphicsEnvironmentImpl.closePath();
        graphicsEnvironmentImpl.fill();
        graphicsEnvironmentImpl.stroke();
    }
}
